@@ -1,10 +1,11 @@
-// Copyright 2021 Martin Dosch.
+// Copyright Martin Dosch.
 // Use of this source code is governed by the BSD-2-clause
 // license that can be found in the LICENSE file.
 
 package main
 
 import (
+	"context"
 	"crypto/tls"
 	"encoding/xml"
 	"fmt"
@@ -20,9 +21,34 @@ import (
 )
 
 const (
-	version = "0.3.4"
+	version = "0.3.7"
 )
 
+// Created with https://github.com/miku/zek
+type StreamError struct {
+	XMLName xml.Name `xml:"stream"`
+	Text    string   `xml:",chardata"`
+	Version string   `xml:"version,attr"`
+	Stream  string   `xml:"stream,attr"`
+	Db      string   `xml:"db,attr"`
+	Lang    string   `xml:"lang,attr"`
+	From    string   `xml:"from,attr"`
+	Xmlns   string   `xml:"xmlns,attr"`
+	ID      string   `xml:"id,attr"`
+	To      string   `xml:"to,attr"`
+	Error   struct {
+		Chardata        string `xml:",chardata"`
+		PolicyViolation struct {
+			Text  string `xml:",chardata"`
+			Xmlns string `xml:"xmlns,attr"`
+		} `xml:"policy-violation"`
+		Text struct {
+			Text  string `xml:",chardata"`
+			Xmlns string `xml:"xmlns,attr"`
+		} `xml:"text"`
+	} `xml:"error"`
+}
+
 var (
 	// statusOK will print [OK] in green color.
 	statusOK = "[\033[32mOK\033[00m]"
@@ -109,12 +135,6 @@ func main() {
 		DoT:      *flagDoT,
 	}
 
-	// Lookup CNAME
-	target, err := c.LookupCNAME(server[0])
-	if err != nil {
-		log.Fatal(err)
-	}
-
 	// Timeout
 	timeout := time.Duration(*flagTimeout) * time.Second
 
@@ -143,7 +163,7 @@ func main() {
 	}
 
 	if *flagClient {
-		clientRecords, err := c.LookupClient(target)
+		clientRecords, err := c.LookupClient(server[0])
 		if err != nil && len(clientRecords) == 0 {
 			fmt.Println(err)
 			if *flagFallback && *flagTest {
@@ -152,17 +172,17 @@ func main() {
 				clientRecords = []xmppsrv.SRV{
 					{
 						Type:   "xmpp-client",
-						Target: target,
+						Target: server[0],
 						Port:   5222,
 					},
 					{
 						Type:   "xmpps-client",
-						Target: target,
+						Target: server[0],
 						Port:   5223,
 					},
 					{
 						Type:   "xmpps-client",
-						Target: target,
+						Target: server[0],
 						Port:   443,
 					},
 				}
@@ -176,7 +196,7 @@ func main() {
 		if *flagClient {
 			fmt.Println()
 		}
-		serverRecords, err := c.LookupServer(target)
+		serverRecords, err := c.LookupServer(server[0])
 		if err != nil && len(serverRecords) == 0 {
 			fmt.Println(err)
 			if *flagFallback && *flagTest {
@@ -185,12 +205,12 @@ func main() {
 				serverRecords = []xmppsrv.SRV{
 					{
 						Type:   "xmpp-server",
-						Target: target,
+						Target: server[0],
 						Port:   5269,
 					},
 					{
 						Type:   "xmpps-server",
-						Target: target,
+						Target: server[0],
 						Port:   5270,
 					},
 				}
@@ -202,7 +222,8 @@ func main() {
 }
 
 func checkRecord(records []xmppsrv.SRV, verbose bool, ipv4 bool, ipv6 bool, test bool,
-	tlsConfig *tls.Config, timeout time.Duration) {
+	tlsConfig *tls.Config, timeout time.Duration,
+) {
 	for count, record := range records {
 		if count > 0 {
 			fmt.Println()
@@ -221,7 +242,8 @@ func printRecord(record xmppsrv.SRV) {
 }
 
 func printIP(record xmppsrv.SRV, ipv4 bool, ipv6 bool, test bool,
-	tlsConfig *tls.Config, timeout time.Duration) {
+	tlsConfig *tls.Config, timeout time.Duration,
+) {
 	addresses, err := net.LookupIP(record.Target)
 	switch {
 	case err != nil:
@@ -248,18 +270,18 @@ func printIP(record xmppsrv.SRV, ipv4 bool, ipv6 bool, test bool,
 				if err == nil {
 					switch record.Type {
 					case "xmpp-client":
-						startTLS("client", c, tlsConfig)
+						startTLS("client", c, tlsConfig, timeout)
 						c.Close()
 					case "xmpps-client":
 						tlsConfig.NextProtos = []string{"xmpp-client"}
-						directTLS("client", c, tlsConfig)
+						directTLS("client", c, tlsConfig, timeout)
 						c.Close()
 					case "xmpp-server":
-						startTLS("server", c, tlsConfig)
+						startTLS("server", c, tlsConfig, timeout)
 						c.Close()
 					case "xmpps-server":
 						tlsConfig.NextProtos = []string{"xmpp-server"}
-						directTLS("server", c, tlsConfig)
+						directTLS("server", c, tlsConfig, timeout)
 						c.Close()
 					default:
 						c.Close()
@@ -281,7 +303,7 @@ func connectionTest(server string, transport string, timeout time.Duration) (net
 	return c, err
 }
 
-func startTLS(recordType string, c net.Conn, tlsConfig *tls.Config) {
+func startTLS(recordType string, c net.Conn, tlsConfig *tls.Config, timeout time.Duration) {
 	// Created with https://github.com/miku/zek
 	type Proceed struct {
 		XMLName xml.Name `xml:"proceed"`
@@ -295,8 +317,9 @@ func startTLS(recordType string, c net.Conn, tlsConfig *tls.Config) {
 		Xmlns   string   `xml:"xmlns,attr"`
 	}
 	var (
-		serverProceed Proceed
-		serverFailure Failure
+		serverProceed     Proceed
+		serverFailure     Failure
+		serverStreamError StreamError
 	)
 	var from string
 	if recordType == "server" {
@@ -318,7 +341,15 @@ func startTLS(recordType string, c net.Conn, tlsConfig *tls.Config) {
 	if err != nil {
 		fmt.Println("StartTLS:", statusNOK)
 		if err.Error() == "EOF" {
-			fmt.Println("Server sent EOF.")
+			// TODO: Properly check for all possible server errors
+			// instead of only checking for "policy-violation".
+			err = xml.Unmarshal(buf, &serverStreamError)
+			if err == nil {
+				fmt.Println("Server sent policy-violation:",
+					serverStreamError.Error.Text.Text)
+			} else {
+				fmt.Println("Server sent EOF.")
+			}
 		} else {
 			fmt.Println(err)
 		}
@@ -342,7 +373,13 @@ func startTLS(recordType string, c net.Conn, tlsConfig *tls.Config) {
 		if err != nil {
 			fmt.Println("StartTLS:", statusNOK)
 			if err.Error() == "EOF" {
-				fmt.Println("Server sent EOF.")
+				err = xml.Unmarshal(buf, &serverStreamError)
+				if err == nil {
+					fmt.Println("Server sent policy-violation:",
+						serverStreamError.Error.Text.Text)
+				} else {
+					fmt.Println("Server sent EOF.")
+				}
 			} else {
 				fmt.Println(err)
 			}
@@ -357,13 +394,20 @@ func startTLS(recordType string, c net.Conn, tlsConfig *tls.Config) {
 		}
 	}
 	d := tls.Client(c, tlsConfig)
-	err = d.Handshake()
+	ctx := context.Background()
+	ctx, cancel := context.WithTimeout(ctx, timeout)
+	defer cancel()
+	err = d.HandshakeContext(ctx)
 	if err != nil {
-		if err.Error() == "EOF" {
+		switch err.Error() {
+		case "EOF":
 			fmt.Println("StartTLS:", statusNOK)
 			fmt.Println("Received EOF during handshake.")
-		} else {
+		case "context deadline exceeded":
 			fmt.Println("StartTLS:", statusNOK)
+			fmt.Println("Timeout during handshake.")
+		default:
+			fmt.Println("STartTLS:", statusNOK)
 			fmt.Println(err)
 		}
 	} else {
@@ -373,14 +417,22 @@ func startTLS(recordType string, c net.Conn, tlsConfig *tls.Config) {
 	}
 }
 
-func directTLS(recordType string, conn net.Conn, tlsConfig *tls.Config) {
+func directTLS(recordType string, conn net.Conn, tlsConfig *tls.Config, timeout time.Duration) {
+	var serverStreamError StreamError
 	c := tls.Client(conn, tlsConfig)
-	err := c.Handshake()
+	ctx := context.Background()
+	ctx, cancel := context.WithTimeout(ctx, timeout)
+	defer cancel()
+	err := c.HandshakeContext(ctx)
 	if err != nil {
-		if err.Error() == "EOF" {
+		switch err.Error() {
+		case "EOF":
 			fmt.Println("TLS:", statusNOK)
 			fmt.Println("Received EOF during handshake.")
-		} else {
+		case "context deadline exceeded":
+			fmt.Println("TLS:", statusNOK)
+			fmt.Println("Timeout during handshake.")
+		default:
 			fmt.Println("TLS:", statusNOK)
 			fmt.Println(err)
 		}
@@ -405,7 +457,13 @@ func directTLS(recordType string, conn net.Conn, tlsConfig *tls.Config) {
 		if err != nil {
 			fmt.Println("TLS:", statusNOK)
 			if err.Error() == "EOF" {
-				fmt.Println("Server sent EOF.")
+				err = xml.Unmarshal(buf, &serverStreamError)
+				if err == nil {
+					fmt.Println("Server sent policy-violation:",
+						serverStreamError.Error.Text.Text)
+				} else {
+					fmt.Println("Server sent EOF.")
+				}
 			} else {
 				fmt.Println(err)
 			}
